
! Copyright 2014 Fabio Cascioli, Jonas Lindert, Philipp Maierhoefer, Stefano Pozzorini
!
! This file is part of OpenLoops.
!
! OpenLoops is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! OpenLoops is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with OpenLoops.  If not, see <http://www.gnu.org/licenses/>.


module openloops
  use KIND_TYPES, only: DREALKIND, MaxParticles
  use, intrinsic :: iso_c_binding, only: c_ptr, c_null_ptr, c_char, c_int, c_double, c_null_char
  use ol_init, only: set_init_error_fatal, get_error, &
    & set_parameter, get_parameter, parameters_flush, cleanup
  implicit none
  private
  ! from module init_iu
  public :: set_init_error_fatal, get_error
  public :: set_parameter, get_parameter, parameters_flush
  ! process interface
  public :: n_external, amplitudetype, phase_space_point, register_process, start, finish
  public :: evaluate_tree, evaluate_cc, evaluate_sc, evaluate_full
  public :: evaluate_loop, evaluate_loop2, evaluate_ct, evaluate_pt
  ! used in BLHA interface
  public :: rval_size

  type process_handle
    integer :: n_particles = 0
    integer :: max_point = -1
    integer :: tensor_rank = -1
    integer, allocatable :: permutation(:)
    type(c_ptr) :: library_handle = c_null_ptr
    integer :: amplitude_type ! 1=Tree, 2=ccTree, 3=scTree, 4=scTree_polvect, 11=Loop, 12=LoopInduced
    integer :: content = 0 ! bitwise: 2^0=tree, 2^1=loop, 2^2=loop2, 2^3=pt
    procedure(), pointer, nopass :: set_permutation => null()
    procedure(), pointer, nopass :: tree => null()
    procedure(), pointer, nopass :: loop => null()
    procedure(), pointer, nopass :: ct => null()
    procedure(), pointer, nopass :: pt => null()
    procedure(), pointer, nopass :: rambo => null()
  end type process_handle

  ! process handle array
  integer, save :: last_process_id = 0
  type(process_handle), save, allocatable :: process_handles(:)

  type flag
    integer :: InterfaceVersion ! 1=BLHA1, 2=BLHA2
    integer :: Model ! 1=SMdiag
    integer :: CorrectionType ! 1=QCD, 2=EW
    integer :: IRregularization ! 1=CDR
    integer :: EWrenormalisation ! 1=alpha0
    integer :: AmplitudeType ! 1=Tree, 2=ccTree, 3=scTree, 4=scTree_polvect, 11=Loop, 12=LoopInduced
    integer :: PoleCheck ! 0/1
    character(100) :: answer_file_name ! Answer file as string
  end type flag

  type(flag) flags

#if __APPLE__
  character(len=5), parameter :: dynlib_extension='dylib'
#else
  character(len=2), parameter :: dynlib_extension='so'
#endif



  contains


  pure function rval_size(n_part, amp_type)
    implicit none
    integer :: rval_size
    integer, intent(in) :: n_part, amp_type
    select case (amp_type)
      case  (1, 12) ! Tree, LoopInduced
        rval_size = 1
      case (11) ! Loop
        rval_size = 4
      case (2) ! ccTreeo
        rval_size = (n_part*(n_part-1))/2
      case (3) ! scTree
        rval_size = 2*n_part*n_part
      case (4) ! scTree_polvect
        rval_size = n_part
    end select
  end function rval_size



  function get_process_handle(lib, proc, perm, content, amptype)
    ! [in] lib: a shared library handle
    ! [in] proc: a full process name, '<lib>_<subproc>_<id>_'
    ! [in] perm: integer array with the crossing
    ! [in] content: integer with binary tags for tree, loop, loop2, pt
    ! [in] amptype: integer to specify BLHA matrix element type
    ! return process handle of type process_handle
    ! note: error handling is done in dlsym
    use ol_dlfcn, only: dlsym
    implicit none
    type(c_ptr), intent(in) :: lib
    character(len=*), intent(in) :: proc
    integer, intent(in) :: perm(:), content, amptype
    type(process_handle) :: get_process_handle
    real(DREALKIND) :: masses(MaxParticles)
    procedure(), pointer :: tmp_fun
    allocate(get_process_handle%permutation(size(perm)))
    get_process_handle%permutation = perm
    get_process_handle%library_handle = lib
    get_process_handle%set_permutation => dlsym(lib, "ol_f_set_permutation_" // trim(proc))
    get_process_handle%rambo => dlsym(lib, "ol_f_rambo_" // trim(proc))
    get_process_handle%amplitude_type = amptype
    get_process_handle%tree => dlsym(lib, "ol_f_amp2_" // trim(proc))
    get_process_handle%loop => dlsym(lib, "ol_f_vamp2_" // trim(proc))
    get_process_handle%ct => dlsym(lib, "ol_f_ctamp2_" // trim(proc))
    get_process_handle%pt => dlsym(lib, "ol_f_ptamp2_" // trim(proc))
    get_process_handle%content = content
    ! number of external particles and highest tensor rank
    tmp_fun => dlsym(lib, "ol_f_n_external_" // trim(proc))
    call tmp_fun(get_process_handle%n_particles)
    if (btest(content, 1)) then
      tmp_fun => dlsym(lib, "ol_f_max_point_" // trim(proc))
      call tmp_fun(get_process_handle%max_point)
      tmp_fun => dlsym(lib, "ol_f_tensor_rank_" // trim(proc))
      call tmp_fun(get_process_handle%tensor_rank)
    end if
    ! check correct size of the permutation
    if (get_process_handle%n_particles /= size(perm)) then
      write(*,*) '[OpenLoops] error: registered process with wrong size of particle permutation'
      stop
    end if
  end function get_process_handle


  function register_process_lib(libname, proc, perm, content, amptype)
    ! [in] libname: name of the process library
    ! [in] proc: a full process name, '<lib>_<subproc>_<id>_'
    ! [in] perm: integer array with the crossing
    ! [in] content: integer with binary tags for tree, loop, loop2, pt
    ! [in] amptype: integer to specify BLHA matrix element type
    ! return (integer) process id to be used in OLP_EvalSubProcess
    use ol_dlfcn, only: dlopen, RTLD_LAZY
    use ol_loop_parameters_decl_/**/DREALKIND, only: maxpoint, maxrank
    implicit none
    character(len=*), intent(in) :: libname
    character(len=*), intent(in) :: proc
    integer, intent(in) :: perm(:), content, amptype
    type(c_ptr) :: lib
    integer :: register_process_lib
    type(process_handle), allocatable :: process_handles_bak(:)
    if (.not. allocated(process_handles)) then
      allocate(process_handles(1))
    end if
    if (last_process_id == size(process_handles)) then
      allocate(process_handles_bak(last_process_id))
      process_handles_bak = process_handles
      deallocate(process_handles)
      allocate(process_handles(2*last_process_id))
      process_handles(1:last_process_id) = process_handles_bak
      deallocate(process_handles_bak)
    end if
    lib = dlopen(libname, RTLD_LAZY, 2)
    last_process_id = last_process_id + 1
    process_handles(last_process_id) = get_process_handle(lib, proc, perm, content, amptype)
    if (maxpoint < process_handles(last_process_id)%max_point) then
      maxpoint = process_handles(last_process_id)%max_point
    end if
    if (maxrank < process_handles(last_process_id)%tensor_rank) then
      maxrank = process_handles(last_process_id)%tensor_rank
    end if
    register_process_lib = last_process_id
  end function register_process_lib



  subroutine unregister_processes()
    ! Close all process libraries and nullify process handles.
    use ol_dlfcn, only: dlclose
    implicit none
    integer :: id
    do id = 1, last_process_id
      call dlclose(process_handles(id)%library_handle)
      process_handles(id)%n_particles = 0
      process_handles(id)%content = 0
      deallocate(process_handles(id)%permutation)
      process_handles(id)%library_handle = c_null_ptr
      process_handles(id)%set_permutation => null()
      process_handles(id)%tree => null()
      process_handles(id)%loop => null()
      process_handles(id)%ct => null()
      process_handles(id)%pt => null()
    end do
    if (allocated(process_handles)) deallocate(process_handles)
    last_process_id = 0
  end subroutine unregister_processes



  function register_process(process, amptype)
    ! process: string with format 2->n-2
    ! amptype: integer 1,2,3,4,11,12
    ! return (integer) process id to be used in evaluate_process
    use KIND_TYPES, only: MaxParticles
    use ol_parameters_decl_/**/DREALKIND, only: &
      & install_path, pid_string, rME, rMM, rML, rMU, rMD, rMC, rMS, rMB, &
      & leadingcolour, allowed_libs, coupling_QCD, coupling_EW, order_ew, order_qcd, &
      & approximation, ckmorder, flavour_mapping_on, tmp_dir, verbose
    use ol_loop_parameters_decl_/**/DREALKIND, only: nf, nc
    use iso_fortran_env, only: iostat_end
    use ol_iso_c_utilities, only: to_lowercase
#ifdef USE_IFORT
    use ifport, only: system
#endif
    implicit none
    character(len=*), intent(in) :: process
    integer, intent(in) :: amptype
    integer :: register_process

    integer, parameter :: max_length = 255
    character(len=4) :: loops_flags
    character(len=4) :: loops_specification
    character(len=max_length) :: libname
    character(len=max_length) :: inp
    character(len=max_length) :: outp
    character(len=max_length) :: tmp
    character(len=max_length) :: subprocnum
    character(len=max_length) :: libhandle
    character(len=max_length) :: proc, procunmapped
    character(len=max_length) :: librarypath
    character(len=max_length) :: libfilename
    character(len=max_length) :: infofilename
    character(len=500) :: grepstring
    character(len=500) :: finalgrepstring
    character(len=500) :: lineinfo
    character(len=max_length) :: perm_str, mapping_str
    character(len=3) :: pvalstr
    character :: coupling_order(2)
    character :: infoME, infoMM, infoML, infoMU, infoMD, infoMS, infoMC, infoMB
    character(len=max_length) :: infoAPPROX
    integer :: lib_content
    integer :: infoEWorder(0:1), infoQCDorder(0:1)
    integer :: infoLeadingColour, infoNF, infoNC, infoCKMORDER
    integer :: ierrparam, i
    integer :: gf, gf2, readok, ierrp, ierrg
    integer :: opend_gf2
    integer :: tmpi
    integer :: found
    integer :: libid
    integer :: next
    integer :: ext(MaxParticles)
    logical :: proclib_exists
    integer, allocatable :: perm(:)
    integer :: librarytype

    loops_flags = "tlsp" ! used in this order to set content bits
    call parameters_flush() ! make sure that pid_string is set
    register_process = -1
    gf = 97
    gf2 = 96
    opend_gf2 = 0
    mapping_str = ""

    ! check that proclib exists
#ifdef USE_GFORTRAN
    inquire(file=trim(install_path)//"/proclib/.", exist=proclib_exists)
#endif
#ifdef USE_IFORT
    inquire(directory=trim(install_path)//"/proclib", exist=proclib_exists)
#endif
    if (.not. proclib_exists) then
      print *, "[OpenLoops] register_process: proclib folder not found, check install_path."
      return
    end if

    ! split: process->(in,out)
    next = 0
    ext = 0
    inp = adjustl(trim(process(1:index(process, "->")-1)))
    outp = adjustl(process(index(process, "->")+2:len(process)))

    ! split initial state and check that we have a 2 -> n production process
    tmp = adjustl(inp(1:index(inp, " ")))
    next = next + 1
    read(tmp,*) ext(next)
    tmp = adjustl(inp(index(inp, " "):))
    if (index(tmp, " ") == 1) then
      print *, "[OpenLoops] register_process: invalid argument: " // trim(process)
      return
    else
      next = next + 1
      read(tmp,*) ext(next)
    end if

    ! split final state
    FSloop: do
      tmp = adjustl(outp(1:index(outp, " ")))
      outp = adjustl(outp(index(outp, " "):))
      if (index(tmp, " ") == 1) then
        exit FSloop !no more final state particles
      else
        next = next + 1
        read(tmp,*) tmpi
        ext(next) = tmpi
      end if
    end do FSloop

    ! flavour mapping
    if (flavour_mapping_on == 1) then
      if (verbose > 1) then
        print*, "[OpenLoops] Using flavour mapping. Original process: ", ext(1:next)
      end if
      call flavour_mapping(ext(1:next))
      if (verbose > 1) then
        print*, "[OpenLoops] Using flavour mapping. Mapped process:   ", ext(1:next)
      end if
    end if

    ! charge conjugate final state particles
    do i=3,next
      ext(i)=charge_conj(ext(i))
    end do

    ! determine normal ordering
    allocate (perm(next))
    call normal_order(ext(1:next), perm, proc)
    if (proc == "") return

    librarytype = 0
    50 continue ! continue from here for a flavour-mapped process or different library classes
    infofilename = "" ! set if following a mapping

    ! find corresponding .info file
    grepstring = 'grep -sH " ' // trim(proc) // ' " ' // trim(install_path) // '/proclib/'
    select case (amptype)
      case (1,2,3,4) ! tree-like
        if (librarytype == 0) then
          loops_specification = "t"
        else if (librarytype == 1) then
          loops_specification = "lt"
        else if (librarytype == 2) then
          loops_specification = "lpt"
        else if (librarytype == 3) then
          loops_specification = "lst"
        else if (librarytype == 4) then
          loops_specification = "lpst"
        else
          print *, "[OpenLoops] register_process: process not found: " // trim(process) // " (" // trim(proc) // ")"
          return
        end if
      case (11) ! loop
        if (librarytype == 0) then
          loops_specification = "lt"
        else if (librarytype == 1) then
          loops_specification = "l"
        else if (librarytype == 2) then
          loops_specification = "lp"
        else if (librarytype == 3) then
          loops_specification = "ls"
        else if (librarytype == 4) then
          loops_specification = "lst"
        else if (librarytype == 5) then
          loops_specification = "lpt"
        else if (librarytype == 6) then
          loops_specification = "lps"
        else if (librarytype == 7) then
          loops_specification = "lpst"
        else
          print *, "[OpenLoops] register_process: process not found: " // trim(process) // " (" // trim(proc) // ")"
          return
        end if
      case (12) ! loop-induced
        if (librarytype == 0) then
          loops_specification = "ls"
        else if (librarytype == 1) then
          loops_specification = "lst"
        else if (librarytype == 2) then
          loops_specification = "lps"
        else if (librarytype == 3) then
          loops_specification = "lpst"
        else
          print *, "[OpenLoops] register_process: process not found: " // trim(process) // " (" // trim(proc) // ")"
          return
        end if
      case default
        print *, "[OpenLoops] register_process: amplitude type not supported:", amptype
    end select
    finalgrepstring = trim(grepstring) // '*_' // trim(loops_specification) // '.info ' // &
                    & '> ' // trim(tmp_dir) // '/OL_output_' // trim(pid_string) // '.tmp'
    ierrg = system(finalgrepstring)

    if (ierrg /= 0) then
      librarytype = librarytype + 1
      go to 50
    end if

    open(gf, file=trim(tmp_dir) // '/OL_output_' // trim(pid_string) // '.tmp', &
       & status = "old", iostat=readok)
    if (readok /= 0) then
      print *, "[OpenLoops] register_process: can't open temporary file."
      return
    end if

    lib_content = 0
    do i = 1, len(loops_flags)
      if (index(loops_specification, loops_flags(i:i)) > 0) lib_content = ibset(lib_content, i-1)
    end do

    InfoReadLoop: do
      read (gf, '(A)', iostat=readok )  lineinfo

      if (readok /= 0) then ! EOF -> exit
        if (readok == iostat_end) then
          exit InfoReadLoop
        else
          print *, "[OpenLoops] register_process: error reading temporary file."
          return
        end if
      end if

      lineinfo = trim(lineinfo)
      if (len_trim(lineinfo) == 0) then
        cycle InfoReadLoop ! strip empty lines
      end if


      !check if library is "allowed"
      libname= lineinfo(index(lineinfo, ':')+1:index(lineinfo, ' '))
      if (allowed_libs /= '' .and. index(allowed_libs, trim(libname)) == 0) then
        cycle InfoReadLoop
      end if

      ! follow mapping
      if(index(lineinfo, 'map=') /= 0) then
            procunmapped = proc
        proc = lineinfo(index(lineinfo, 'map=')+4:index(lineinfo, 'map=')+3+len_trim(proc))
        mapping_str = lineinfo(index(lineinfo, '[')+1:index(lineinfo, ']')-1)
        !check for MB etc.
        call readInfo(lineinfo, 'MB', infoMB)
        if ((infoMB == "0" .and. rMB /= 0) ) then
          if (verbose > 1) then
            print*, "[OpenLoops] Not following massless b-mapping ", trim(procunmapped), " --> ", trim(proc) , ": MB /= 0."
          end if
          proc = procunmapped
          cycle InfoReadLoop
        end if
        if (verbose > 1) then
          print*, "[OpenLoops] Following info-file mapping: ",trim(procunmapped), &
         &        " --> ", trim(lineinfo(index(lineinfo, 'map=')+4:)), "."
        end if
        !map permutation
        if(mapping_str /= '') then
          call map_permutation(perm,mapping_str)
          mapping_str = " (mapped from " // trim(procunmapped) // ")"
        end if
        infofilename= lineinfo(index(lineinfo, 'libopenloops'):index(lineinfo, '.info')+4)
        grepstring = 'grep -sH " ' // trim(proc) // ' " ' // trim(install_path) // '/proclib/'
        finalgrepstring = trim(grepstring) // trim(infofilename) // &
          & ' > ' // trim(tmp_dir) // '/OL_output_mapping_' // trim(pid_string) // '.tmp'
        ierrg = system(finalgrepstring)
        if (ierrg /= 0) then
          print *, "[OpenLoops] register_process: inconsistent mapping in " // trim(process)
          return
        end if
        open(gf2, file=trim(tmp_dir) // '/OL_output_mapping_' // trim(pid_string) // '.tmp', &
           & status="OLD", iostat=readok)
        if (readok /= 0) then
          print *, "[OpenLoops] register_process: can't open temporary file."
          return
        else
          opend_gf2 = 1
        end if
      end if

      51 continue
      if (infofilename /= "") then
        read(gf2, '(A)', iostat=readok) lineinfo
        if (readok /= 0) then ! EOF -> exit
          if (readok == iostat_end) then
            close(gf2)
            ierrg = system("rm -f " // trim(tmp_dir) // "/OL_output_mapping_" // trim(pid_string) // ".tmp")
            infofilename = ""
            proc = procunmapped
            cycle InfoReadLoop
          else
            print *, "[OpenLoops] register_process: error reading temporary file."
            return
          end if
        end if
        if (len_trim(lineinfo) == 0) then
          go to 51
        end if
      end if

      ! get library filename
      libfilename = lineinfo(index(lineinfo, 'libopenloops'):index(lineinfo, '.info')) // dynlib_extension
      ! remove path
      lineinfo = lineinfo(index(lineinfo, '.info')+6:)

      subprocnum = lineinfo(index(lineinfo, trim(proc))+len_trim(proc)+1:index(lineinfo, trim(proc))+len_trim(proc)+2)

      call readInfoCoupling(lineinfo, 'QCD', infoQCDorder)
      call readInfoCoupling(lineinfo, 'EW', infoEWorder)
      call readInfo(lineinfo, 'ME', infoME)
      call readInfo(lineinfo, 'MM', infoMM)
      call readInfo(lineinfo, 'ML', infoML)
      call readInfo(lineinfo, 'MU', infoMU)
      call readInfo(lineinfo, 'MD', infoMD)
      call readInfo(lineinfo, 'MS', infoMS)
      call readInfo(lineinfo, 'MC', infoMC)
      call readInfo(lineinfo, 'MB', infoMB)
      call readInfo(lineinfo, 'APPROX', infoAPPROX)
      call readInfoInt(lineinfo, 'CKMORDER', infoCKMorder)
      call readInfoInt(lineinfo, 'nc', infoNC)
      call readInfoInt(lineinfo, 'nf', infoNF)
      call readInfoInt(lineinfo, 'LeadingColour', infoLeadingColour)

      !set couplings with either order_ew/order_qcd (the one which is kept fixed)
      if (order_ew /= -1 .and. order_qcd == -1 ) then
        coupling_EW(0)=order_ew
        coupling_QCD(1)=1
      else if (order_ew == -1 .and. order_qcd /= -1) then
        coupling_QCD(0)=order_qcd
        coupling_ew(1)=1
      else if (order_ew /= -1 .and.  order_qcd /= -1) then
        print*, "[OpenLoops] Error: order_ew and order_qcd can not be set at the same time!"
        goto 52
      end if

      if ((infoEWorder(0) == coupling_EW(0) .or. coupling_EW(0) == -1) &
        & .and. (infoEWorder(1) == coupling_EW(1) .or. coupling_EW(1) == -1) &
        & .and. (infoQCDorder(0) == coupling_QCD(0) .or. coupling_QCD(0) == -1) &
        & .and. (infoQCDorder(1) == coupling_QCD(1) .or. coupling_QCD(1) == -1) &
        & .and. infoLeadingColour == leadingcolour &
        & .and. infoNC == nc &
        & .and. infoNF == nf &
        & .and. infoCKMorder == ckmorder &
        & .and. ((infoME /= "0" .and. rME /= 0) .or. (rME == 0)) &
        & .and. ((infoMM /= "0" .and. rMM /= 0) .or. (rMM == 0)) &
        & .and. ((infoML /= "0" .and. rML /= 0) .or. (rML == 0)) &
        & .and. ((infoMU /= "0" .and. rMU /= 0) .or. (rMU == 0)) &
        & .and. ((infoMD /= "0" .and. rMD /= 0) .or. (rMD == 0)) &
        & .and. ((infoMS /= "0" .and. rMS /= 0) .or. (rMS == 0)) &
        & .and. ((infoMC /= "0" .and. rMC /= 0) .or. (rMC == 0)) &
        & .and. ((infoMB /= "0" .and. rMB /= 0) .or. (rMB == 0)) &
        & .and. (trim(infoAPPROX)==trim(approximation)) &
        & ) then
         if (verbose > 1) then
          print*, "[OpenLoops] Parameters match info-file for process ", trim(proc), " in library ", trim(libfilename)
        end if
        found = 1 ! found correct library -> exit
        exit
      else if (infofilename /= "") then
        if (verbose > 1) then
          print*, "[OpenLoops] Parameters do not match info-file for process ", trim(proc), " in library ", trim(libfilename)
        end if
        found = 0
        go to 51
      else
        if (verbose > 1) then
          print*, "[OpenLoops] Parameters do not match info-file for process ", trim(proc), " in library ", trim(libfilename)
        end if
        found = 0
      end if

    end do InfoReadLoop
    52 continue
    close(gf)
    ierrg = system("rm -f " // trim(tmp_dir) // "/OL_output_" // trim(pid_string) // ".tmp")
    if (opend_gf2 .eq. 1) then
      close(gf2)
      ierrg = system("rm -f " // trim(tmp_dir) // "/OL_output_mapping_" // trim(pid_string) // ".tmp")
    end if

    ! load library
    if (found == 1) then
      libfilename = trim(install_path) // '/proclib/' // libfilename
      libhandle = trim(to_lowercase(libname)) // "_" // trim(proc) // "_" // trim(subprocnum)
      register_process = register_process_lib(libfilename, libhandle, perm, lib_content, amptype)

      perm_str = "["
      do i = 1, next - 1
        write(pvalstr,'(I2)') perm(i)
        perm_str = trim(perm_str) // trim(adjustl(pvalstr)) // ","
      end do
      write(pvalstr,'(I2)') perm(i)
      perm_str = trim(perm_str) // trim(adjustl(pvalstr)) // "]"
!        write(lineout,'(A,A,I4,A,A,A)') trim(line),  " | 1 " , libid, " # ", trim(libhandle), trim(perm_str)

      if (verbose > 0) then
        print *, "[OpenLoops] Loaded process ", trim(process), " : " , trim(libhandle), &
                    &  trim(perm_str),  trim(mapping_str), "."
      end if

    else
      librarytype = librarytype + 1
      go to 50
    end if

    !deallocate
    deallocate(perm)

  contains

  function charge_conj(x)
      ! determine charge conjugate.
      implicit none
      integer, intent(in)  :: x
      integer :: charge_conj
      select case(x)
        case(0, 21, 22, 23, 25)
          charge_conj = x
        case default
          charge_conj = -x
      end select
  end function charge_conj

    subroutine map_permutation(perm, mapping_str)
      implicit none
      character(len=30), intent(inout) :: mapping_str
      integer, intent(inout) :: perm(:)
      character(len=2) :: str
      integer, allocatable :: mapping(:), perm_tmp(:)
      integer :: i, x
      allocate (mapping(size(perm)))
      allocate (perm_tmp(size(perm)))

      mapping_str=trim(mapping_str) // ','
      do i = 1, size(mapping)
        str = adjustl(mapping_str(0:index(mapping_str,',')-1))
        read(str(2:), '(i10)') mapping(i)
        mapping_str = mapping_str(index(mapping_str,',')+1:len_trim(mapping_str))
      end do
      !map permutation
      do i = 1, size(mapping)
        perm_tmp(i) = mapping(perm(i))
      end do
      perm = perm_tmp

      deallocate(mapping)
      deallocate(perm_tmp)
    end subroutine map_permutation


    subroutine readInfo(lineinfo, var, res)
      implicit none
      character(len=*), intent(in) :: lineinfo
      character(len=*), intent(in) :: var
      character(len=*), intent(out) :: res
      if (index(lineinfo, var//'=') /= 0) then
        res = lineinfo(index(lineinfo, var//'=')+len_trim(var)+1: &
              & index(lineinfo, var//'=')+index(lineinfo(index(lineinfo, var//'='):),' ')-1 )
      else
        res = ""
      end if
    end subroutine readInfo

    subroutine readInfoInt(lineinfo, var, res)
      implicit none
      character(len=*), intent(in) :: lineinfo
      character(len=*), intent(in) :: var
      character(len=10) :: restemp
      integer, intent(out) :: res
      if (index(lineinfo, var//'=') /= 0) then
        restemp = lineinfo(index(lineinfo, var//'=')+len_trim(var)+1: &
            & index(lineinfo, var//'=')+index(lineinfo(index(lineinfo, var//'='):),' ')-1)
      else
        restemp = "0"
      end if
      read(restemp,'(I2)') res
    end subroutine readInfoInt

    subroutine readInfoCoupling(lineinfo, var, res)
      implicit none
      character(len=*), intent(in) :: lineinfo
      character(len=*), intent(in) :: var
      character(len=1) :: restemp(0:1)
      integer, intent(out) :: res(0:1)
      if (index(lineinfo, var//'=') /= 0) then
        restemp(0) =  lineinfo(index(lineinfo, trim(var))+len_trim(var)+1:index(lineinfo, trim(var))+len_trim(var)+2)
        restemp(1) =  lineinfo(index(lineinfo, trim(var))+len_trim(var)+3:index(lineinfo, trim(var))+len_trim(var)+4)
      else
        restemp(0) = "0"
        restemp(1) = "0"
      end if
      read(restemp(0),'(I2)') res(0)
      read(restemp(1),'(I2)') res(1)
    end subroutine readInfoCoupling

  end function register_process


  subroutine flavour_mapping(ext)
!   Lepton & quark flavour mapping
!   Concept (taken from old Sherpa interface):
!   (1) Given a final state, determine the four (anti)lepton/neutrino
!       multiplicities in the given process:
!       a_nubar, a_nu, a_lbar, a_l
!   (2) Compute the discriminant N[i] as
!       N[i] = Ngen - (i+1) + Ngen*(a_nubar + a_nu*Nmax + a_lbar*Nmax^2 + a_l*Nmax^3)
!       where Nmax should be chosen such that a_...<=Nmax.
!       In practice one can safely set Nmax=10 and it will work for any
!       process with <= 20 final-state leptons.
!       It is also convenient to set Ngen=10, although i runs only from 1 to 3.
!   (3) Reassign the lepton generations with a permutation
!       p1 -> 1, p2 -> 2, p3 -> 3  such that  N[p1] > N[p2] > N[p3] */
    implicit none
    integer, intent(inout) :: ext(:)
    integer, allocatable :: new_ext(:)
    integer i, j
    integer Ngen, Nlgen, Nqgen, Nmax
    integer l_gen, nu_gen, l_gen_new, nu_gen_new
    integer d_gen, u_gen, d_gen_new, u_gen_new
    integer a_nu, a_nubar, a_l, a_lbar
    integer a_u, a_ubar, a_d, a_dbar
    integer Nl(3,2), Nq(2,2)
    real(DREALKIND) mtau

    Ngen=10
    Nmax=10

    call get_parameter("tau_mass", mtau)
    if (mtau == 0.) then
      Nlgen = 3
    else
      Nlgen = 2
    end if

    allocate(new_ext(size(ext)))
    new_ext = ext

    !lepton flavour mapping
    do i = 1, Nlgen
      l_gen=9+2*i;
      nu_gen=10+2*i;

      a_nu=count_flav(ext,nu_gen);
      a_nubar=count_flav(ext,-nu_gen);
      a_l=count_flav(ext,l_gen);
      a_lbar=count_flav(ext,-l_gen);

      Nl(i,1)=Ngen-(i+1) + Ngen*(a_nubar+a_nu*Nmax+a_lbar*Nmax*Nmax+a_l*Nmax*Nmax*Nmax)
      Nl(i,2)=i
    end do

    call sort_pair(Nl,Nlgen)

    do i = 1, Nlgen
      l_gen=9+2*Nl(i,2);
      l_gen_new=9+2*i;
      nu_gen=10+2*Nl(i,2);
      nu_gen_new=10+2*i;

      do j = 1, size(ext)
        if (abs(ext(j))==nu_gen) new_ext(j)=sign(nu_gen_new,ext(j))
        if (abs(ext(j))==l_gen)  new_ext(j)=sign(l_gen_new,ext(j))
      end do
    end do

    ext = new_ext

    Nqgen = 2
    !quark flavour mapping
    do i = 1, Nqgen
      d_gen=2*i-1;
      u_gen=2*i;

      a_d=count_flav(ext,d_gen);
      a_dbar=count_flav(ext,-d_gen);
      a_u=count_flav(ext,u_gen);
      a_ubar=count_flav(ext,-u_gen);

      Nq(i,1)=Ngen-(i+1) + Ngen*(a_ubar+a_u*Nmax+a_dbar*Nmax*Nmax+a_d*Nmax*Nmax*Nmax)
      Nq(i,2)=i
    end do

    call sort_pair(Nq,Nqgen)

    do i = 1, 2
      d_gen=2*Nq(i,2)-1;
      d_gen_new=2*i-1;
      u_gen=2*Nq(i,2);
      u_gen_new=2*i;

      do j = 1, size(ext)
        if (abs(ext(j))==u_gen) new_ext(j)=sign(u_gen_new,ext(j))
        if (abs(ext(j))==d_gen) new_ext(j)=sign(d_gen_new,ext(j))
      end do
    end do


    ext = new_ext
    deallocate(new_ext)

  end subroutine


  function count_flav(ext, pid)
  ! count frequency of integer pid in array ext
    implicit none
    integer, intent(in) :: ext(:)
    integer, intent(in) :: pid
    integer :: count_flav
    integer i

    count_flav = 0

    do i = 1, size(ext)
      if (ext(i) == pid) count_flav = count_flav+1
    end do

  end function count_flav


  subroutine sort_pair(a,n)
  ! Simple insertion sort. Sorting descending on the first component of a 2-tuple of length n
    integer, intent(in) :: n
    integer, intent(inout), dimension(:,:) :: a
    integer :: temp(2)
    integer :: i, j

    do i = 2, n
      j = i - 1
      temp(1) = a(i,1)
      temp(2) = a(i,2)
      do while (j>=1 .and. a(j,1)<temp(1))
        a(j+1,1) = a(j,1)
        a(j+1,2) = a(j,2)
        j = j - 1
        if (j==0) exit
      end do
      a(j+1,1) = temp(1)
      a(j+1,2) = temp(2)
    end do
  end subroutine sort_pair


  subroutine normal_order(ext, perm, proc)
      implicit none
      integer, intent(in) :: ext(:)
      integer, intent(out) :: perm(:)
      character(len=*), intent(out) :: proc
      integer :: i,j, normal(31), pos
      character(len=3) :: normalc(31)

      ! define normal ordering and corresponding characters
      normal(  1:10) = [ 12  ,-12  , 14  ,-14  , 16  ,-16  , 11  ,-11  , 13  ,-13  ]
      normalc( 1:10) = ["ne ","nex","nm ","nmx","nl ","nlx","e  ","ex ","m  ","mx "]

      normal( 11:20) = [ 15  ,-15  ,  2  , -2  ,  4  , -4  ,  6  , -6  ,  1  , -1  ]
      normalc(11:20) = ["l  ","lx ","u  ","ux ","c  ","cx ","t  ","tx ","d  ","dx "]

      normal( 21:30) = [  3  , -3  ,  5  , -5  , 25  , 22  , 23  , -24 , 24  , 21  ]
      normalc(21:30) = ["s  ","sx ","b  ","bx ","h  ","a  ","z  ","w  ","wx ","g  "]

      normal( 31:31) = [ 0 ]
      normalc(31:31) = ["g  "]

      perm = 0
      proc = ""

      ! normal order, build string and store permutation
      pos = 1
      do i = 1, size(normal)
        do j = 1, size(ext)
          if (ext(j) == normal(i)) then
            proc = trim(proc) // trim(normalc(i))
            perm(j) = pos
            pos = pos + 1
          end if
        end do
      end do

      if (pos-1 /= size(ext)) then
        print *, "[OpenLoops] normal_order: invalid process specification:", ext
        proc = ""
      end if
  end subroutine normal_order


  function register_process_c(process, amptype) bind(c,name="ol_register_process")
    use ol_iso_c_utilities, only: c_f_string
    implicit none
    character(kind=c_char), dimension(*), intent(in) :: process
    integer(c_int), value :: amptype
    integer(c_int) :: register_process_c
    character(len=:), allocatable :: f_process
    integer :: f_amptype
    f_amptype = amptype
    call c_f_string(process, f_process)
    register_process_c = register_process(f_process, f_amptype)
  end function register_process_c


  pure function amplitudetype(id)
    ! [in] id: a process id
    ! return amptype of type integer
    implicit none
    integer, intent(in) :: id
    integer amplitudetype
    amplitudetype =  process_handles(id)%amplitude_type
  end function amplitudetype


  pure function amplitudetype_c(id) bind(c,name="ol_amplitudetype")
    ! [in] id: a process id
    ! return amptype of type integer
    implicit none
    integer(c_int), intent(in) :: id
    integer(c_int) amplitudetype_c
    amplitudetype_c =  process_handles(id)%amplitude_type
  end function amplitudetype_c


  pure function n_external(id)
    implicit none
    integer, intent(in) :: id
    integer :: n_external
    n_external = process_handles(id)%n_particles
  end function n_external


  function n_external_c(id) bind(c,name="ol_n_external")
    implicit none
    integer(c_int), value :: id
    integer(c_int) :: n_external_c
    n_external_c = process_handles(id)%n_particles
  end function n_external_c


  subroutine phase_space_point(id, sqrt_s, psp)
    implicit none
    integer, intent(in) :: id
    real(DREALKIND), intent(in) :: sqrt_s
    real(DREALKIND), intent(out) :: psp(:,:)
    type(process_handle) :: subprocess
    subprocess = process_handles(id)
    call subprocess%set_permutation(subprocess%permutation)
    call subprocess%rambo(sqrt_s, psp)
  end subroutine phase_space_point


  subroutine phase_space_point_c(id, sqrt_s, pp) bind(c,name="ol_phase_space_point")
    implicit none
    integer(c_int), value :: id
    real(c_double), value :: sqrt_s
    real(c_double), intent(out) :: pp(5*n_external(id))
    integer :: f_id
    real(DREALKIND) :: f_sqrt_s
    real(DREALKIND) :: f_psp(0:3,n_external(id))
    f_id = id
    f_sqrt_s = sqrt_s
    call phase_space_point(f_id, f_sqrt_s, f_psp)
    do f_id = 1, n_external(id)
      pp(5*(f_id-1)+1:5*(f_id-1)+4) = f_psp(0:3,f_id)
      pp(5*f_id) = -1
    end do
  end subroutine phase_space_point_c


  subroutine start() bind(c,name="ol_start")
    implicit none
    call parameters_flush()
  end subroutine


  subroutine finish() bind(c,name="ol_finish")
    implicit none
    call cleanup()
    call unregister_processes()
  end subroutine finish


  subroutine evaluate_tree(id, psp, res)
    implicit none
    integer, intent(in) :: id
    real(DREALKIND), intent(in) :: psp(:,:)
    real(DREALKIND), intent(out) :: res
    real(DREALKIND) :: m2cc(0:n_external(id)*(n_external(id)+1)/2+1)
    type(process_handle) :: subprocess
    subprocess = process_handles(id)
    if (id <= 0 .or. id > last_process_id) then
      write(*,*) '[OpenLoops] evaluate: no registered process with id', id
      stop
    else if (.not. btest(subprocess%content, 0)) then
      write(*,*) '[OpenLoops] evaluate: tree routine not available for process', id
      stop
    end if
    call subprocess%set_permutation(subprocess%permutation)
    call parameters_flush()
    call subprocess%tree(psp, m2cc, 0, &
      & [0._/**/DREALKIND, 0._/**/DREALKIND, 0._/**/DREALKIND, 0._/**/DREALKIND], 1, [0])
    res = m2cc(0)
  end subroutine evaluate_tree


  subroutine evaluate_tree_c(id, pp, res) bind(c,name="ol_evaluate_tree")
    implicit none
    integer(c_int), value :: id
    real(c_double), intent(in) :: pp(5*process_handles(id)%n_particles)
    real(c_double), intent(out) :: res
    integer :: f_id
    real(DREALKIND) :: f_pp(0:4,process_handles(id)%n_particles)
    real(DREALKIND) :: f_res
    f_id = id
    f_pp = reshape(pp, [5,process_handles(id)%n_particles])
    call evaluate_tree(f_id, f_pp(0:3,:), f_res)
    res = f_res
  end subroutine evaluate_tree_c


  subroutine evaluate_cc(id, psp, res)
    implicit none
    integer, intent(in) :: id
    real(DREALKIND), intent(in) :: psp(:,:)
    real(DREALKIND), intent(out) :: res(:)
    type(process_handle) :: subprocess
    real(DREALKIND) :: m2cc(0:n_external(id)*(n_external(id)+1)/2+1)
    integer  :: n_cc, i, j
    subprocess = process_handles(id)
    if (id <= 0 .or. id > last_process_id) then
      write(*,*) '[OpenLoops] evaluate: no registered process with id', id
      stop
    else if (.not. btest(subprocess%content, 0)) then
      write(*,*) '[OpenLoops] evaluate: cc routine not available for process', id
      stop
    end if
    call subprocess%set_permutation(subprocess%permutation)
    n_cc = subprocess%n_particles*(subprocess%n_particles+1)/2
    call parameters_flush()
    call subprocess%tree(psp, m2cc, 0, &
      & [0._/**/DREALKIND, 0._/**/DREALKIND, 0._/**/DREALKIND, 0._/**/DREALKIND], &
      & n_cc, [(i, i = 0, n_cc)])
    do j = 1, subprocess%n_particles - 1
      do i = 0, j - 1
        res(i+j*(j-1)/2+1) = m2cc((j+1)*j/2+i+1)
      end do
    end do
  end subroutine evaluate_cc


  subroutine evaluate_cc_c(id, pp, res) bind(c,name="ol_evaluate_cc")
    implicit none
    integer(c_int), value :: id
    real(c_double), intent(in) :: pp(5*process_handles(id)%n_particles)
    real(c_double), intent(out) :: res(rval_size(process_handles(id)%n_particles,2))
    integer :: f_id
    real(DREALKIND) :: f_pp(0:4,process_handles(id)%n_particles)
    real(DREALKIND) :: f_res(rval_size(process_handles(id)%n_particles,2))
    f_id = id
    f_pp = reshape(pp, [5,process_handles(id)%n_particles])
    call evaluate_cc(f_id, f_pp(0:3,:), f_res)
    res = f_res
  end subroutine evaluate_cc_c


  subroutine evaluate_sc(id, psp, emitter, polvect, res)
    ! Spin correlated matrix elements.
    ! [in] id: process id as set by register_process
    ! [in] psp: phase space point
    ! [in] int emitter: emitter
    ! [in] polvect: polarisation vector
    ! [out] res(n_external): array with results for each spectator j,
    !       res(j) = 1/mom^2 * <emitter,mu|mom^mu cc_ij mom^nu|j,nu>
    implicit none
    integer, intent(in) :: id, emitter
    real(DREALKIND), intent(in) :: psp(:,:), polvect(4)
    real(DREALKIND), intent(out) :: res(:)
    type(process_handle) :: subprocess
    integer :: j, extcombs(n_external(id))
    real(DREALKIND) :: m2sc(0:n_external(id)*(n_external(id)+1)/2+1)
    subprocess = process_handles(id)
    if (id <= 0 .or. id > last_process_id) then
      write(*,*) '[OpenLoops] evaluate: no registered process with id', id
      stop
    else if (.not. btest(subprocess%content, 0)) then
      write(*,*) '[OpenLoops] evaluate: sc routine not available for process', id
      stop
    end if
    do j = 1, subprocess%n_particles
      if (j <= emitter) then
        extcombs(j) = emitter*(emitter-1)/2 + j
      else
        extcombs(j) = j*(j-1)/2 + emitter
      end if
    end do
    call parameters_flush()
    call subprocess%tree(psp, m2sc, emitter, polvect, subprocess%n_particles, extcombs)
    do j = 1, subprocess%n_particles
      res(j) = m2sc(extcombs(j))
    end do
  end subroutine evaluate_sc


  subroutine evaluate_sc_c(id, pp, emitter, polvect, res) bind(c,name="ol_evaluate_sc")
    implicit none
    integer(c_int), value :: id, emitter
    real(c_double), intent(in) :: pp(5*process_handles(id)%n_particles), polvect(4)
    real(c_double), intent(out) :: res(process_handles(id)%n_particles)
    integer :: f_id, f_emitter
    real(DREALKIND) :: f_pp(0:4,process_handles(id)%n_particles), f_polvect(4)
    real(DREALKIND) :: f_res(process_handles(id)%n_particles)
    f_id = id
    f_pp = reshape(pp, [5,process_handles(id)%n_particles])
    f_emitter = emitter
    f_polvect = polvect
    call evaluate_sc(f_id, f_pp(0:3,:), f_emitter, f_polvect, f_res)
    res = f_res
  end subroutine evaluate_sc_c


  subroutine evaluate_full(id, psp, m2l0, m2l1, ir1, m2l2, ir2, acc)
    use ol_stability
    implicit none
    integer, intent(in) :: id
    real(DREALKIND), intent(in) :: psp(:,:)
    real(DREALKIND), intent(out) :: m2l0, m2l1(0:2), ir1(0:2), m2l2(0:4), ir2(0:4)
    real(DREALKIND), intent(out) :: acc
    type(process_handle)  :: subprocess
    subprocess = process_handles(id)
    if (id <= 0 .or. id > last_process_id) then
      write(*,*) '[OpenLoops] evaluate: no registered process with id', id
      stop
    else if (.not. btest(subprocess%content, 1)) then
      write(*,*) '[OpenLoops] evaluate: loop routine not available for process', id
      stop
    end if
    call subprocess%set_permutation(subprocess%permutation)
    call parameters_flush()
    call subprocess%loop(psp, m2l0, m2l1, ir1, m2l2, ir2)
    acc = last_relative_deviation
  end subroutine evaluate_full


  subroutine evaluate_full_c(id, pp, m2l0, m2l1, ir1, m2l2, ir2, acc) bind(c,name="ol_evaluate_full")
    implicit none
    integer(c_int), value :: id
    real(c_double), intent(in) :: pp(5*process_handles(id)%n_particles)
    real(c_double), intent(out) :: m2l0, m2l1(0:2), ir1(0:2), m2l2(0:4), ir2(0:4), acc
    integer :: f_id
    real(DREALKIND) :: f_pp(0:4,process_handles(id)%n_particles)
    real(DREALKIND) :: f_m2l0, f_m2l1(0:2), f_ir1(0:2), f_m2l2(0:4), f_ir2(0:4)
    real(DREALKIND) :: f_acc
    f_id = id
    f_pp = reshape(pp, [5,process_handles(id)%n_particles])
    call evaluate_full(f_id, f_pp(0:3,:), f_m2l0, f_m2l1, f_ir1, f_m2l2, f_ir2, f_acc)
    m2l0 = f_m2l0
    m2l1 = f_m2l1
    ir1  = f_ir1
    m2l2 = f_m2l2
    ir2  = f_ir2
    acc  = f_acc
  end subroutine evaluate_full_c


  subroutine evaluate_loop(id, psp, m2l0, m2l1, acc)
    implicit none
    integer, intent(in)  :: id
    real(DREALKIND), intent(in)  :: psp(:,:)
    real(DREALKIND), intent(out) :: m2l0, m2l1(0:2)
    real(DREALKIND), intent(out) :: acc
    real(DREALKIND) :: ir1(0:2), m2l2(0:4), ir2(0:4)
    call evaluate_full(id, psp, m2l0, m2l1, ir1, m2l2, ir2, acc)
  end subroutine evaluate_loop


  subroutine evaluate_loop_c(id, pp, m2l0, m2l1, acc) bind(c,name="ol_evaluate_loop")
    implicit none
    integer(c_int), value :: id
    real(c_double), intent(in) :: pp(5*process_handles(id)%n_particles)
    real(c_double), intent(out) :: m2l0, m2l1(0:2), acc
    integer :: f_id
    real(DREALKIND) :: f_pp(0:4,process_handles(id)%n_particles)
    real(DREALKIND) :: f_m2l0, f_m2l1(0:2)
    real(DREALKIND) :: f_acc
    f_id = id
    f_pp = reshape(pp, [5,process_handles(id)%n_particles])
    call evaluate_loop(f_id, f_pp(0:3,:), f_m2l0, f_m2l1, f_acc)
    m2l0 = f_m2l0
    m2l1 = f_m2l1
    acc  = f_acc
  end subroutine evaluate_loop_c


  subroutine evaluate_loop2(id, psp, res, acc)
    implicit none
    integer, intent(in)  :: id
    real(DREALKIND), intent(in)  :: psp(:,:)
    real(DREALKIND), intent(out) :: res
    real(DREALKIND), intent(out) :: acc
    real(DREALKIND) :: m2l0, m2l1(0:2), ir1(0:2), m2l2(0:4), ir2(0:4)
    if (.not. btest(process_handles(id)%content, 2)) then
      write(*,*) '[OpenLoops] evaluate: loop routine not available for process', id
      stop
    end if
    call evaluate_full(id, psp, m2l0, m2l1, ir1, m2l2, ir2, acc)
    res = m2l2(0)
  end subroutine evaluate_loop2


  subroutine evaluate_loop2_c(id, pp, res, acc) bind(c,name="ol_evaluate_loop2")
    implicit none
    integer(c_int), value :: id
    real(c_double), intent(in) :: pp(5*process_handles(id)%n_particles)
    real(c_double), intent(out) :: res, acc
    integer :: f_id
    real(DREALKIND) :: f_pp(0:4,process_handles(id)%n_particles)
    real(DREALKIND) :: f_res, f_acc
    f_id = id
    f_pp = reshape(pp, [5,process_handles(id)%n_particles])
    call evaluate_loop2(f_id, f_pp(0:3,:), f_res, f_acc)
    res = f_res
    acc  = f_acc
  end subroutine evaluate_loop2_c


  subroutine evaluate_ct(id, psp, m2l0, m2ct)
    use ol_stability
    implicit none
    integer, intent(in)  :: id
    real(DREALKIND), intent(in)  :: psp(:,:)
    real(DREALKIND), intent(out) :: m2l0, m2ct
    type(process_handle)  :: subprocess
    subprocess = process_handles(id)
    if (id <= 0 .or. id > last_process_id) then
      write(*,*) '[OpenLoops] evaluate: no registered process with id', id
      stop
    else if (.not. btest(subprocess%content, 1)) then
      write(*,*) '[OpenLoops] evaluate: ct routine not available for process', id
      stop
    end if
    call subprocess%set_permutation(subprocess%permutation)
    call parameters_flush()
    call subprocess%ct(psp, m2l0, m2ct)
  end subroutine evaluate_ct


  subroutine evaluate_ct_c(id, pp, m2l0, m2ct) bind(c,name="ol_evaluate_ct")
    implicit none
    integer(c_int), value :: id
    real(c_double), intent(in) :: pp(5*process_handles(id)%n_particles)
    real(c_double), intent(out) :: m2l0, m2ct
    integer :: f_id
    real(DREALKIND) :: f_pp(0:4,process_handles(id)%n_particles)
    real(DREALKIND) :: f_m2l0, f_m2ct
    f_id = id
    f_pp = reshape(pp, [5,process_handles(id)%n_particles])
    call evaluate_ct(f_id, f_pp(0:3,:), f_m2l0, f_m2ct)
    m2l0 = f_m2l0
    m2ct = f_m2ct
  end subroutine evaluate_ct_c


  subroutine evaluate_pt(id, psp, m2l0, m2pt, m2l1)
    use ol_stability
    implicit none
    integer, intent(in) :: id
    real(DREALKIND), intent(in)  :: psp(:,:)
    real(DREALKIND), intent(out) :: m2l0, m2pt, m2l1
    type(process_handle)  :: subprocess
    subprocess = process_handles(id)
    if (id <= 0 .or. id > last_process_id) then
      write(*,*) '[OpenLoops] evaluate: no registered process with id', id
      stop
    else if (.not. btest(subprocess%content, 3)) then
      write(*,*) '[OpenLoops] evaluate: ct routine not available for process', id
      stop
    end if
    call subprocess%set_permutation(subprocess%permutation)
    call parameters_flush()
    call subprocess%pt(psp, m2l0, m2pt, m2l1)
  end subroutine evaluate_pt


  subroutine evaluate_pt_c(id, pp, m2l0, m2pt, m2l1) bind(c,name="ol_evaluate_pt")
    implicit none
    integer(c_int), value :: id
    real(c_double), intent(in) :: pp(5*process_handles(id)%n_particles)
    real(c_double), intent(out) :: m2l0, m2pt, m2l1
    integer :: f_id
    real(DREALKIND) :: f_pp(0:4,process_handles(id)%n_particles)
    real(DREALKIND) :: f_m2l0, f_m2pt, f_m2l1
    f_id = id
    f_pp = reshape(pp, [5,process_handles(id)%n_particles])
    call evaluate_pt(f_id, f_pp(0:3,:), f_m2l0, f_m2pt, f_m2l1)
    m2l0 = f_m2l0
    m2pt = f_m2pt
    m2l1 = f_m2l1
  end subroutine evaluate_pt_c


end module openloops
